#include <stdio.h>
#include <unistd.h>
#include <locale.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>

#include "gen.h"
#include "error.h"
#include "log.h"
#include "time.h"
#include "daytime.h"
#include "mssl.h"
#include "http.h"
#include "ntpd.h"

#define TIME_TIME	1
#define TIME_DAYTIME	2
#define TIME_HTTP	3

#define DEFAULT_SLEEP_INTERVAL 60

int verbose = 0;

void version(void)
{
	printf("OmniSync v" VERSION ", (C) 2007 by folkert@vanheusden.com\n\n");
}

void help(void)
{
	version();

	printf("-m x   mode:\n");
	printf("       time/tcp time/udp\n");
	printf("       daytime/tcp daytime/udp\n");
	printf("       http/https\n");
	printf("-p x   proxy-server (http/https only)\n");
	printf("-h x   host to connect to\n");
	printf("-u x   ntpd shared memory unit\n");
	printf("-f     do not fork\n");
	printf("-v     increase verbosity\n");
	printf("-V     show version & exit\n");
	printf("--help this help\n");
}

int main(int argc, char *argv[])
{
	SSL_CTX *ctx = initialize_ctx();
	int mode = -1, ip_mode = -1, http_mode = -1;
	char *host = NULL, *proxy = NULL;
	int host_port = -1, proxy_port = -1;
	struct shmTime * pst = NULL;
	int unit_nr = 0;
	double fudge_factor = -1;
	int sleep_interval = DEFAULT_SLEEP_INTERVAL;
	int c;
	char do_fork = 1;
	char *dummy = NULL;

	while((c = getopt(argc, argv, "m:p:h:u:fvVh")) != -1)
	{
		switch(c)
		{
		case 'm':
			if (strcasecmp(optarg, "time/tcp") == 0)
			{
				mode = TIME_TIME;
				ip_mode = IP_TCP;
			}
			else if (strcasecmp(optarg, "time/udp") == 0)
			{
				mode = TIME_TIME;
				ip_mode = IP_UDP;
			}
			else if (strcasecmp(optarg, "daytime/tcp") == 0)
			{
				mode = TIME_DAYTIME;
				ip_mode = IP_TCP;
			}
			else if (strcasecmp(optarg, "daytime/udp") == 0)
			{
				mode = TIME_DAYTIME;
				ip_mode = IP_UDP;
			}
			else if (strcasecmp(optarg, "http") == 0)
			{
				mode = TIME_HTTP;
				http_mode = HTTP;
			}
			else if (strcasecmp(optarg, "https") == 0)
			{
				mode = TIME_HTTP;
				http_mode = HTTPS;
			}
			else
			{
				fprintf(stderr, "'%s' is not recognized\n", optarg);
				help();
				return 1;
			}
			break;

		case 'p':
			proxy_port = 8080; 	/* default */
			proxy = strdup(optarg);
			dummy = strchr(proxy, ':');
			if (dummy)
			{
				*dummy = 0x00;
				proxy_port = atoi(dummy + 1);
			}
			break;

		case 'h':
			host = strdup(optarg);
			host_port = 80;
			dummy = strchr(host, ':');
			if (dummy)
			{
				*dummy = 0x00;
				host_port = atoi(dummy + 1);
			}
			break;

		case 'f':
			do_fork = 0;
			break;

		case 'u':
			unit_nr = atoi(optarg);
			if (unit_nr < 0)
				error_exit("-u requires a positive value\n");
			if (unit_nr > 3)
				error_exit("NTPd normally supports only 4 shared memory devices, still continuing though\n");
			break;

		case 'v':
			verbose++;
			break;

		case 'V':
			version();
			return 0;

		default:
			help();
			return 1;
		}
	}

	if (!do_fork || verbose)
		version();

	if (mode == -1)
	{
		fprintf(stderr, "No mode given: use the '-m' flag\n");
		return 1;
	}

	if (host == NULL)
	{
		fprintf(stderr, "No host to connect to given, use the '-h' flag\n");
		return 1;
	}

	if ((mode == TIME_TIME || mode == TIME_DAYTIME) && proxy != NULL)
	{
		fprintf(stderr, "time/daytime does not support a proxy server\n");
		return 1;
	}
	else if (mode == TIME_HTTP && ip_mode != -1 && ip_mode != IP_TCP)
	{
		fprintf(stderr, "http(s) only works via TCP\n");
		return 1;
	}

	if (setlocale(LC_ALL, "C") == NULL)
		error_exit("Failed to set locale to 'C' - which is required for correct daytime-conversion.");

	pst = get_shm_pointer(unit_nr);
	if (!pst)
		error_exit("Failed to connect to NTP daemon");
	
	if (do_fork)
	{
		if (daemon(-1, -1) == -1)
			error_exit("Failed to become daemon process");
	}

	for(;;)
	{
		int sleep_left = sleep_interval;
		int rc = -1;
		double ts_start_recv, ts_measurement;

		if (mode == TIME_TIME)
		{
			rc = gtime(host, ip_mode, &ts_start_recv, &ts_measurement);
		}
		else if (mode == TIME_DAYTIME)
		{
			rc = daytime(host, ip_mode, &ts_start_recv, &ts_measurement);
		}
		else if (mode == TIME_HTTP)
		{
			rc = httptime(host, host_port, proxy, proxy_port, http_mode, ctx, &ts_start_recv, &ts_measurement);
		}
		else
			error_exit("unknown mode %d", mode);

		if (rc == 0)
		{
			// FIXME send to ntp, only if not too large difference

			if (submit_to_ntpd(pst, ts_start_recv, ts_measurement, fudge_factor))
			{
				// count did not fetches
				printf("did not fetch\n");
			}

			if (verbose)
				printf("%f\n", ts_measurement - ts_start_recv);
		}

		do
		{
			sleep_left = sleep(sleep_left);

			if (sleep_left > 0 && verbose > 1)
				dolog(LOG_DEBUG, "Early return from sleep (%d) seconds left", sleep_left);
		}
		while(sleep_left > 0);
	}

	return 0;
}
